"""
.. module:: util.py

--- NOTE: All of this is pretty much deprecated by the transport.HTTPTransport object

"""

import urllib
import re
import codecs
from twisted.internet import defer
from twisted.internet import reactor
from zope.interface import implements
from twisted import version as twisted_version
from twisted.web import client
from twisted.web.http import PotentialDataLoss
from twisted.web.http_headers import Headers
from twisted.web.iweb import IBodyProducer
from twisted.internet.protocol import Protocol
from twisted.web._newclient import ResponseDone
from StringIO import StringIO

if twisted_version.major >= 12:
    Agent = client.Agent
else:
    class Agent(client.Agent):
        """
        client.Agent does not provide the ability to set a timeout.
        This is an Agent with a timeout.
        """

        def __init__(self, reactor,
                           contextFactory=None,
                           connectTimeout=240):
            self._reactor = reactor
            self._contextFactory = contextFactory
            self.connectTimeout = connectTimeout


        def _connect(self, scheme, host, port):
            """ connect method with a timeout. """

            cc = client.ClientCreator(self._reactor, self._protocol)
            if scheme == 'http':
                d = cc.connectTCP(host, port, timeout=self.connectTimeout)
            elif scheme == 'https':
                raise Exception('HTTPS not supported')
                #d = cc.connectSSL(host, port, self._wrapContextFactory(host, port),
                #                  timeout=self.timeout)
            else:
                d = defer.fail(SchemeNotSupported(
                        "Unsupported scheme: %r" % (scheme,)))
            return d

class StringProducer(object):
    """
    Body producer for t.w.c.Agent
    """
    implements(IBodyProducer)

    def __init__(self, body):
        self.body = body
        self.length = len(body)

    def startProducing(self, consumer):
        return defer.maybeDeferred(consumer.write, self.body)

    def pauseProducing(self):
        pass

    def stopProducing(self):
        pass


class ResponseReceiver(Protocol):
    """
    Assembles HTTP response from return stream.
    """

    def __init__(self, deferred):
        self.writer = codecs.getwriter("utf_8")(StringIO())
        self.deferred = deferred

    def dataReceived(self, bytes):
        self.writer.write(bytes)

    def connectionLost(self, reason):
        if reason.check(ResponseDone) or reason.check(PotentialDataLoss):
            self.deferred.callback(self.writer.getvalue())
        else:
            self.deferred.errback(reason)

def getPageWithHeaders(contextFactory=None,
                      scheme='http', host='127.0.0.1', port=80, path='/',
                      timeout=240, *args, **kwargs):
    """Download a web page as a string.

    :returns: twisted.web.client.Agent deferred.

    Download a page. Return a deferred, which will callback with a
    page (as a string) or errback with a description of the error.

    See t.w.c.Agent to see what extra args can be passed.
    """

    def cb_recv_resp(response):
        d_resp_recvd = defer.Deferred()
        if response.length:
            response.deliverBody(ResponseReceiver(d_resp_recvd))
            return d_resp_recvd.addCallback(cb_process_resp, response)
        else:
            return cb_process_resp("", response)

    def cb_process_resp(body, response):
        _headers = {"http_code": response.code}

        for header in response.headers.getAllRawHeaders():
            _headers[header[0].lower()] = header[1][0]

        return _headers, body

    url = str("http://%s:%d%s" % (host, port, path))

    if not "headers" in kwargs:
        kwargs["headers"] = {}
    else:
        for header in kwargs["headers"]:
            kwargs["headers"][header] = [kwargs["headers"][header]]

    if not "method" in kwargs:
        kwargs["method"] == "GET"

    if "postdata" in kwargs:
        body = StringProducer(kwargs["postdata"])
    else:
        body = None

    d = Agent(reactor, connectTimeout=timeout).request(kwargs["method"],
                                      url,
                                      Headers(kwargs["headers"]),
                                      body
                                  )

    d.addCallback(cb_recv_resp)
    return d

def flatten_js(js_func):
    """Flatten a JavaScript function into a single line.

    :returns: String.

    Take a JavaScript function and strip out all of the
    newlines so that it is suitable for passing into a
    Riak map/reduce phase.
    """

    return "".join(js_func.split("\n"))

def get_value(key, array, default_value):
    """
    Overkill for array.get(key, default_value).
    Likely due to porting from another language.
    """
    if (key in array):
        return array[key]
    else:
        return default_value

def build_rest_path(client,
                    bucket=None, key=None, spec=None, params=None,
                    prefix=None):
    """
    Given a RiakClient, RiakBucket, Key, LinkSpec, and Params,
    construct and return a URL.
    """
    # Build 'http://hostname:port/prefix/bucket'
    path = ''
    path += '/' + (prefix or client._prefix)

    # Add '.../bucket'
    if bucket:
        path += '/' + urllib.quote_plus(bucket._name)

    # Add '.../key'
    if bucket and key:
        path += '/' + urllib.quote_plus(key)

    # Add query parameters.
    if (params != None):
        s = ''
        for key in params.keys():
            if (s != ''):
                s += '&'
            s += (urllib.quote_plus(key) + '=' +
                  urllib.quote_plus(str(params[key])))
        path += '?' + s

    # Return.
    return client._host, client._port, path

def http_request_deferred(method, host, port, path,
                          headers=None, obj=''):
    """
    Given a Method, URL, Headers, and Body, perform an HTTP request,
    and return deferred.
    :returns: deferred
    """
    if headers == None:
        headers = {}

    return getPageWithHeaders(contextFactory=None,
                              scheme='http',
                              host=host,
                              method=method,
                              port=port,
                              path=path,
                              headers=headers,
                              postdata=obj,
                              timeout=20)

def build_headers(headers):
    """
    Build up the header string.
    """
    headers1 = []
    for key in headers.keys():
        headers1.append('%s: %s' % (key, headers[key]))
    return headers1

def parse_http_headers(headers):
    """
    Parse an HTTP Header string into dictionary of response headers.
    """
    result = {}
    fields = headers.split("\n")
    for field in fields:
        matches = re.match("([^:]+):(.+)", field)
        if (matches == None):
            continue
        key = matches.group(1).lower()
        value = matches.group(2).strip()
        if (key in result.keys()):
            if  isinstance(result[key], list):
                result[key].append(value)
            else:
                result[key] = [result[key]].append(value)
        else:
            result[key] = value
    return result





from zope.interface import implements

from twisted.internet import defer, reactor, protocol
from twisted.web.client import Agent
from twisted.web.http_headers import Headers
from twisted.web.iweb import IBodyProducer

from distutils.version import StrictVersion

import urllib
import re, csv
from cStringIO import StringIO

from xml.etree import ElementTree

# MD_ resources
from riakasaurus.metadata import *

from riakasaurus.riak_index_entry import RiakIndexEntry
from riakasaurus.mapreduce import RiakLink

MAX_LINK_HEADER_SIZE = 8192 - 8

versions = {
    1: StrictVersion("1.0.0"),
    1.1: StrictVersion("1.1.0"),
    1.2: StrictVersion("1.2.0")
    }

class FeatureDetection(object):
    _s_version = None

    def _server_version(self):
        """
        Gets the server version from the server. To be implemented by
        the individual transport class.
        :rtype string
        """
        raise NotImplementedError

    @defer.inlineCallbacks
    def phaseless_mapred(self):
        """
        Whether MapReduce requests can be submitted without phases.
        :rtype bool
        """
        d = yield self.server_version() 
        defer.returnValue(d >= versions[1.1])

    @defer.inlineCallbacks
    def pb_indexes(self):
        """
        Whether secondary index queries are supported over Protocol
        Buffers

        :rtype bool
        """
        d = yield self.server_version() 
        defer.returnValue(d >= versions[1.2])

    @defer.inlineCallbacks
    def pb_search(self):
        """
        Whether search queries are supported over Protocol Buffers
        :rtype bool
        """
        d = yield self.server_version() 
        defer.returnValue(d >= versions[1.2])

    @defer.inlineCallbacks
    def pb_conditionals(self):
        """
        Whether conditional fetch/store semantics are supported over
        Protocol Buffers
        :rtype bool
        """
        d = yield self.server_version() 
        defer.returnValue(d >= versions[1])

    @defer.inlineCallbacks
    def quorum_controls(self):
        """
        Whether additional quorums and FSM controls are available,
        e.g. primary quorums, basic_quorum, notfound_ok
        :rtype bool
        """
        d = yield self.server_version() 
        defer.returnValue(d >= versions[1])

    @defer.inlineCallbacks
    def tombstone_vclocks(self):
        """
        Whether 'not found' responses might include vclocks
        :rtype bool
        """
        d = yield self.server_version() 
        defer.returnValue(d >= versions[1])

    @defer.inlineCallbacks
    def pb_head(self):
        """
        Whether partial-fetches (vclock and metadata only) are
        supported over Protocol Buffers
        :rtype bool
        """
        d = yield self.server_version() 
        defer.returnValue(d >= versions[1])

    @defer.inlineCallbacks
    def server_version(self):
        if not self._s_version:
            self._s_version = yield self._server_version()

        defer.returnValue(StrictVersion(self._s_version))

class BodyReceiver(protocol.Protocol):
    """ Simple buffering consumer for body objects """
    def __init__(self, finished):
        self.finished = finished 
        self.buffer = StringIO()

    def dataReceived(self, buffer):
        self.buffer.write(buffer)

    def connectionLost(self, reason):
        self.buffer.seek(0)
        self.finished.callback(self.buffer)

class StringProducer(object):
    """
    Body producer for t.w.c.Agent
    """
    implements(IBodyProducer)

    def __init__(self, body):
        self.body = body
        self.length = len(body)

    def startProducing(self, consumer):
        consumer.write(self.body)
        return defer.succeed(None)

    def pauseProducing(self):
        pass

    def stopProducing(self):
        pass

class HTTPTransport(FeatureDetection):
    """ HTTP Transport for Riak """
    def __init__(self, client):
        self._prefix = client._prefix
        self.host = client._host
        self.port = client._port
        self.client = client
        self._client_id = None

    def http_response(self, response):
        def haveBody(body):
            headers = {"http_code": response.code}
            for key, val in response.headers.getAllRawHeaders():
                headers[key.lower()] = val[0]
            
            return headers, body.read()

        if response.length:
            d = defer.Deferred()
            response.deliverBody(BodyReceiver(d))
            return d.addCallback(haveBody)
        else:
            return haveBody(StringIO(""))

    def http_request(self, method, path, headers={}, body=None):
        url = "http://%s:%s%s" % (self.host, self.port, path)

        h = {}
        for k, v in headers.items():
            if not isinstance(v, list):
                h[k] = [v]
            else:
                h[k] = v 

        if body:
            bodyProducer = StringProducer(body)
        else:
            bodyProducer = None

        return Agent(reactor).request(
                method, str(url), Headers(h), bodyProducer
            ).addCallback(self.http_response)

    def build_rest_path(self, bucket=None, key=None, params=None, prefix=None) :
        """
        Given a RiakClient, RiakBucket, Key, LinkSpec, and Params,
        construct and return a URL.
        """
        # Build 'http://hostname:port/prefix/bucket'
        path = ''
        path += '/' + (prefix or self._prefix)

        # Add '.../bucket'
        if bucket is not None:
            path += '/' + urllib.quote_plus(bucket._name)

        # Add '.../key'
        if key is not None:
            path += '/' + urllib.quote_plus(key)

        # Add query parameters.
        if params is not None:
            s = ''
            for key in params.keys():
                if params[key] is not None:
                    if s != '': s += '&'
                    s += urllib.quote_plus(key) + '=' + urllib.quote_plus(str(params[key]))
            path += '?' + s

        # Return.
        return path

    def decodeJson(self, s):
        return self.client.get_decoder('application/json')(s)

    def encodeJson(self, s):
        return self.client.get_encoder('application/json')(s)

    @defer.inlineCallbacks
    def get_keys(self, bucket):
        params = {'props' : 'True', 'keys' : 'true'}
        url = self.build_rest_path(bucket, params=params)
        

        headers, encoded_props = yield self.http_request('GET', url)

        if headers['http_code'] == 200:
            props = self.decodeJson(encoded_props)
        else:
            raise Exception('Error getting bucket properties.') 

        defer.returnValue(props['keys'])

    @defer.inlineCallbacks
    def set_bucket_props(self, bucket, props):
        """
        Set bucket properties
        """ 
        url = self.build_rest_path(bucket)
        headers = {'Content-Type': 'application/json'}
        content = self.encodeJson({'props': props})

        #Run the request...
        headers, response = yield self.http_request('PUT', url, headers, content)

        # Handle the response...
        if (response is None):
            raise Exception('Error setting bucket properties.')

        # Check the response value...
        status = headers['http_code']

        if (status != 204):
            raise Exception('Error setting bucket properties.')

        defer.returnValue(response)

    def set_client_id(self, client_id):
        self._client_id = client_id

    def get_client_id(self):
        return self._client_id

    @defer.inlineCallbacks
    def ping(self):
        """
        Check server is alive over HTTP
        """
        response = yield self.http_request('GET', '/ping')
        res = (response is not None) and (response[1] == 'OK')
        defer.returnValue(res)

    @defer.inlineCallbacks
    def stats(self):
        """
        Gets performance statistics and server information
        """
        # TODO: use resource detection
        response = yield self.http_request('GET', '/stats', {'Accept':'application/json'})
        if response[0]['http_code'] is 200:
            defer.returnValue(self.decodeJson(response[1]))
        else:
            defer.returnValue(None)

    # FeatureDetection API - private
    @defer.inlineCallbacks
    def _server_version(self):
        stats = yield self.stats()
        if stats is not None:
            defer.returnValue(stats['riak_kv_version'])
        # If stats is disabled, we can't assume the Riak version
        # is >= 1.1. However, we can assume the new URL scheme is
        # at least version 1.0
        elif 'riak_kv_wm_buckets' in self.get_resources():
            defer.returnValue("1.0.0")
        else:
            defer.returnValue("0.14.0")
   
    @defer.inlineCallbacks
    def get_resources(self):
        """
        Gets a JSON mapping of server-side resource names to paths
        :rtype dict
        """
        response = yield self.http_request('GET', '/', {'Accept':'application/json'})
        if response[0]['http_status'] is 200:
            defer.returnValue(self.decodeJson(response[1]))
        else:
            defer.returnValue({})

    @defer.inlineCallbacks
    def get(self, robj, r = None, pr = None, vtag = None) :
        """
        Get a bucket/key from the server
        """
        # We could detect quorum_controls here but HTTP ignores
        # unknown flags/params.
        params = {'r' : r, 'pr': pr}
        if vtag is not None:
            params['vtag'] = vtag
        url = self.build_rest_path(robj.get_bucket(), robj.get_key(),
                                   params=params)
        response = yield self.http_request('GET', url)
        defer.returnValue(
            self.parse_body(response, [200, 300, 404])
        )

    def put(self, robj, w = None, dw = None, pw = None, return_body = True, if_none_match=False):
        """
        Serialize put request and deserialize response
        """
        # We could detect quorum_controls here but HTTP ignores
        # unknown flags/params.
        params = {'returnbody' : str(return_body).lower(), 'w' : w, 'dw' : dw, 'pw' : pw }
        url = self.build_rest_path(bucket=robj.get_bucket(), key=robj.get_key(),
                                   params=params)
        headers = self.build_put_headers(robj)
        # TODO: use a more general 'prevent_stale_writes' semantics,
        # which is a superset of the if_none_match semantics.
        if if_none_match:
            headers["If-None-Match"] = "*"
        content = robj.get_encoded_data()
        return self.do_put(url, headers, content, return_body, key=robj.get_key())

    @defer.inlineCallbacks
    def do_put(self, url, headers, content, return_body=False, key=None):
        if key is None:
            response = yield self.http_request('POST', url, headers, content)
        else:
            response = yield self.http_request('PUT', url, headers, content)

        if return_body:
            defer.returnValue(self.parse_body(response, [200, 201, 300]))
        else:
            self.check_http_code(response, [204])
            defer.returnValue(None)

    @defer.inlineCallbacks
    def put_new(self, robj, w=None, dw=None, pw=None, return_body=True, if_none_match=False):
        """Put a new object into the Riak store, returning its (new) key."""
        # We could detect quorum_controls here but HTTP ignores
        # unknown flags/params.
        params = {'returnbody' : str(return_body).lower(), 'w' : w, 'dw' : dw, 'pw' : pw}
        url = self.build_rest_path(bucket=robj.get_bucket(), params=params)
        headers = self.build_put_headers(robj)
        # TODO: use a more general 'prevent_stale_writes' semantics,
        # which is a superset of the if_none_match semantics.
        if if_none_match:
            headers["If-None-Match"] = "*"
        content = robj.get_encoded_data()
        response = yield self.http_request('POST', url, headers, content)
        location = response[0]['location']
        idx = location.rindex('/')
        key = location[idx+1:]
        if return_body:
            vclock, [(metadata, data)] = self.parse_body(response, [201])
            defer.returnValue((key, vclock, metadata))
        else:
            self.check_http_code(response, [201])
            defer.returnValue((key, None, None))

    @defer.inlineCallbacks
    def delete(self, robj, rw=None, r = None, w = None, dw = None, pr = None, pw = None):
        """
        Delete an object.
        """
        # We could detect quorum_controls here but HTTP ignores
        # unknown flags/params.
        params = {'rw' : rw, 'r': r, 'w': w, 'dw': dw, 'pr': pr, 'pw': pw}
        headers = {}
        url = self.build_rest_path(robj.get_bucket(), robj.get_key(),
                                   params=params)
        ts = yield self.tombstone_vclocks()
        if ts and robj.vclock() is not None:
            headers['X-Riak-Vclock'] = robj.vclock()
        response = yield self.http_request('DELETE', url, headers)
        self.check_http_code(response, [204, 404])
        defer.returnValue(self)

    @defer.inlineCallbacks
    def get_buckets(self):
        """
        Fetch a list of all buckets
        """
        params = {'buckets': 'true'}
        url = self.build_rest_path(None, params=params)
        response = yield self.http_request('GET', url)

        headers, encoded_props = response[0:2]
        if headers['http_code'] == 200:
            props = self.decodeJson(encoded_props)
        else:
            raise Exception('Error getting buckets.')

        defer.returnValue(props['buckets'])

    @defer.inlineCallbacks
    def get_bucket_props(self, bucket):
        """
        Get properties for a bucket
        """
        # Run the request...
        params = {'props' : 'True', 'keys' : 'False'}
        url = self.build_rest_path(bucket, params=params)
        response = yield self.http_request('GET', url)

        headers = response[0]
        encoded_props = response[1]
        if headers['http_code'] == 200:
            props = self.decodeJson(encoded_props)
            defer.returnValue(props['props'])
        else:
            raise Exception('Error getting bucket properties.')

    @defer.inlineCallbacks
    def mapred(self, inputs, query, timeout=None):
        """
        Run a MapReduce query.
        """
        plm = yield self.phaseless_mapred()
        if not plm and (query is None or len(query) is 0):
            raise Exception('Phase-less MapReduce is not supported by this Riak node')

        # Construct the job, optionally set the timeout...
        job = {'inputs':inputs, 'query':query}
        if timeout is not None:
            job['timeout'] = timeout

        content = self.encodeJson(job)

        # Do the request...
        url = "/" + self.client._mapred_prefix
        headers = {'Content-Type': 'application/json'}
        response = yield self.http_request('POST', url, headers, content)

        # Make sure the expected status code came back...
        status = response[0]['http_code']
        if status != 200:
            raise Exception('Error running MapReduce operation. Headers: %s Body: %s' %
                            (repr(response[0]),repr(response[1])))

        result = self.decodeJson(response[1])
        defer.returnValue(result)

    def get_index(self, bucket, index, startkey, endkey=None):
        """
        Performs a secondary index query.
        """
        # TODO: use resource detection
        segments = ["buckets", bucket, "index", index, str(startkey)]
        if endkey:
            segments.append(str(endkey))
        uri = '/%s' % ('/'.join(segments))
        headers, data = response = self.get_request(uri)
        self.check_http_code(response, [200])
        jsonData = self.decodeJson(data)
        return jsonData[u'keys'][:]

    def search(self, index, query, **params):
        """
        Performs a search query.
        """
        if index is None:
            index = 'search'

        options = {'q':query, 'wt':'json'}
        if 'op' in params:
            op = params.pop('op')
            options['q.op'] = op

        options.update(params)
        # TODO: use resource detection
        uri = "/solr/%s/select" % index
        headers, data = response = self.get_request(uri, options)
        self.check_http_code(response, [200])
        if 'json' in headers['content-type']:
            results = self.decodeJson(data)
            return self._normalize_json_search_response(results)
        elif 'xml' in headers['content-type']:
            return self._normalize_xml_search_response(data)
        else:
            raise ValueError("Could not decode search response")

    def check_http_code(self, response, expected_statuses):
        status = response[0]['http_code']
        if not status in expected_statuses:
            m = 'Expected status ' + str(expected_statuses) + ', received ' + str(status) + ' : ' + response[1]
            raise Exception(m)

    def parse_body(self, response, expected_statuses):
        """
        Given the output of RiakUtils.http_request and a list of
        statuses, populate the object. Only for use by the Riak client
        library.
        @return self
        """
        # If no response given, then return.
        if response is None:
            return self

        # Make sure expected code came back
        self.check_http_code(response, expected_statuses)

        # Update the object...
        headers = response[0]
        data = response[1]
        status = headers['http_code']

        # Check if the server is down(status==0)
        if not status:
            ### we need the host/port that was used.
            m = 'Could not contact Riak Server: http://$HOST:$PORT !'
            raise RiakError(m)

        # If 404(Not Found), then clear the object.
        if status == 404:
            return None

        # If 300(Siblings), then return the list of siblings
        elif status == 300:
            # Parse and get rid of 'Siblings:' string in element 0
            siblings = data.strip().split('\n')
            siblings.pop(0)
            return siblings

        # Parse the headers...
        vclock = None
        metadata = {MD_USERMETA: {}, MD_INDEX: []}
        links = []
        for header, value in headers.iteritems():
            if header == 'content-type':
                metadata[MD_CTYPE] = value
            elif header == 'charset':
                metadata[MD_CHARSET] = value
            elif header == 'content-encoding':
                metadata[MD_CTYPE] = value
            elif header == 'etag':
                metadata[MD_VTAG] = value
            elif header =='link':
                self.parse_links(links, headers['link'])
            elif header == 'last-modified':
                metadata[MD_LASTMOD] = value
            elif header.startswith('x-riak-meta-'):
                metadata[MD_USERMETA][header.replace('x-riak-meta-', '')] = value
            elif header.startswith('x-riak-index-'):
                field = header.replace('x-riak-index-', '')
                reader = csv.reader([value], skipinitialspace=True)
                for line in reader:
                    for token in line:
                        rie = RiakIndexEntry(field, token)
                        metadata[MD_INDEX].append(rie)
            elif header == 'x-riak-vclock':
                vclock = value
            elif header == 'x-riak-deleted':
                metadata[MD_DELETED] = True
        if links:
            metadata[MD_LINKS] = links

        return vclock, [(metadata, data)]

    def to_link_header(self, link):
        """
        Convert this RiakLink object to a link header string. Used internally.
        """
        header = ''
        header += '</'
        header += self._prefix + '/'
        header += urllib.quote_plus(link.get_bucket()) + '/'
        header += urllib.quote_plus(link.get_key()) + '>; riaktag="'
        header += urllib.quote_plus(link.get_tag()) + '"'
        return header

    def parse_links(self, links, linkHeaders):
        """
        Private.
        @return self
        """
        for linkHeader in linkHeaders.strip().split(','):
            linkHeader = linkHeader.strip()
            matches = re.match("</([^/]+)/([^/]+)/([^/]+)>; ?riaktag=\"([^\']+)\"", linkHeader) or \
                re.match("</(buckets)/([^/]+)/keys/([^/]+)>; ?riaktag=\"([^\']+)\"", linkHeader)
            if matches is not None:
                link = RiakLink(urllib.unquote_plus(matches.group(2)),
                                urllib.unquote_plus(matches.group(3)),
                                urllib.unquote_plus(matches.group(4)))
                links.append(link)
        return self

    def add_links_for_riak_object(self, robject, headers):
        links = robject.get_links()
        if links:
            current_header = ''
            for link in links:
                header = self.to_link_header(link)
                if len(current_header + header) > MAX_LINK_HEADER_SIZE:
                    headers.add('Link', current_header)
                    current_header = ''

                if current_header != '': header = ', ' + header
                current_header += header

            headers['Link'] = current_header

        return headers

    def get_request(self, uri=None, params=None):
        url = self.build_rest_path(bucket=None, params=params, prefix=uri)

        return self.http_request('GET', url)

    def store_file(self, key, content_type="application/octet-stream", content=None):
        url = self.build_rest_path(prefix='luwak', key=key)
        headers = {'Content-Type' : content_type,
                   'X-Riak-ClientId' : self._client_id}

        return self.do_put(url, headers, content, key=key)

    @defer.inlineCallbacks
    def get_file(self, key):
        url = self.build_rest_path(prefix='luwak', key=key)
        response = yield self.http_request('GET', url)
        result = self.parse_body(response, [200, 300, 404])
        if result is not None:
            (vclock, data) = result
            (headers, body) = data.pop()
            defer.returnValue(body)

    @defer.inlineCallbacks
    def delete_file(self, key):
        url = self.build_rest_path(prefix='luwak', key=key)
        response = yield self.http_request('DELETE', url)
        self.parse_body(response, [204, 404])

    def post_request(self, uri=None, body=None, params=None, content_type="application/json"):
        uri = self.build_rest_path(prefix=uri, params=params)
        return self.http_request('POST', uri, {'Content-Type': content_type}, body)

    # Utility functions used by Riak library.

    def build_rest_path(self, bucket=None, key=None, params=None, prefix=None) :
        """
        Given a RiakClient, RiakBucket, Key, LinkSpec, and Params,
        construct and return a URL.
        """
        # Build 'http://hostname:port/prefix/bucket'
        path = ''
        path += '/' + (prefix or self._prefix)

        # Add '.../bucket'
        if bucket is not None:
            path += '/' + urllib.quote_plus(bucket._name)

        # Add '.../key'
        if key is not None:
            path += '/' + urllib.quote_plus(key)

        # Add query parameters.
        if params is not None:
            s = ''
            for key in params.keys():
                if params[key] is not None:
                    if s != '': s += '&'
                    s += urllib.quote_plus(key) + '=' + urllib.quote_plus(str(params[key]))
            path += '?' + s

        # Return.
        return path

    def build_put_headers(self, robj):
        """Build the headers for a POST/PUT request."""

        # Construct the headers...
        headers = {'Accept' : 'text/plain, */*; q=0.5',
                   'Content-Type' : robj.get_content_type(),
                   'X-Riak-ClientId' : self._client_id}

        # Add the vclock if it exists...
        if robj.vclock() is not None:
            headers['X-Riak-Vclock'] = robj.vclock()

        # Create the header from metadata
        links = self.add_links_for_riak_object(robj, headers)

        for key, value in robj.get_usermeta().iteritems():
            headers['X-Riak-Meta-%s' % key] = value

        for rie in robj.get_indexes():
            key = 'X-Riak-Index-%s' % rie.get_field()
            if key in headers:
                headers[key] += ", " + rie.get_value()
            else:
                headers[key] = rie.get_value()

        return headers

    def _normalize_json_search_response(self, json):
        """
        Normalizes a JSON search response so that PB and HTTP have the
        same return value
        """
        result = {}
        if u'response' in json:
            result['num_found'] = json[u'response'][u'numFound']
            result['max_score'] = float(json[u'response'][u'maxScore'])
            docs = []
            for doc in json[u'response'][u'docs']:
                resdoc = {u'id': doc[u'id']}
                if u'fields' in doc:
                    for k, v in doc[u'fields'].iteritems():
                        resdoc[k] = v
                docs.append(resdoc)
            result['docs'] = docs
        return result

    def _normalize_xml_search_response(self, xml):
        """
        Normalizes an XML search response so that PB and HTTP have the
        same return value
        """
        target = XMLSearchResult()
        parser = ElementTree.XMLParser(target = target)
        parser.feed(xml)
        return parser.close()

    @classmethod
    def build_headers(cls, headers):
        return ['%s: %s' % (header, value) for header, value in headers.iteritems()]

    @classmethod
    def parse_http_headers(cls, headers) :
        """
        Parse an HTTP Header string into an asssociative array of
        response headers.
        """
        retVal = {}
        fields = headers.split("\n")
        for field in fields:
            matches = re.match("([^:]+):(.+)", field)
            if matches is None: continue
            key = matches.group(1).lower()
            value = matches.group(2).strip()
            if key in retVal.keys():
                if  isinstance(retVal[key], list):
                    retVal[key].append(value)
                else:
                    retVal[key] = [retVal[key]].append(value)
            else:
                retVal[key] = value
        return retVal

class XMLSearchResult(object):
    # Match tags that are document fields
    fieldtags = ['str', 'int', 'date']

    def __init__(self):
        # Results
        self.num_found = 0
        self.max_score = 0.0
        self.docs = []

        # Parser state
        self.currdoc = None
        self.currfield = None
        self.currvalue = None

    def start(self, tag, attrib):
        if tag == 'result':
            self.num_found = int(attrib['numFound'])
            self.max_score = float(attrib['maxScore'])
        elif tag == 'doc':
            self.currdoc = {}
        elif tag in self.fieldtags and self.currdoc is not None:
            self.currfield = attrib['name']

    def end(self, tag):
        if tag == 'doc' and self.currdoc is not None:
            self.docs.append(self.currdoc)
            self.currdoc = None
        elif tag in self.fieldtags and self.currdoc is not None:
            if tag == 'int':
                self.currvalue = int(self.currvalue)
            self.currdoc[self.currfield] = self.currvalue
            self.currfield = None
            self.currvalue = None

    def data(self, data):
        if self.currfield:
            # riak_solr_output adds NL + 6 spaces
            data = data.rstrip()
            if self.currvalue:
                self.currvalue += data
            else:
                self.currvalue = data

    def close(self):
        return {'num_found':self.num_found,
                'max_score':self.max_score,
                'docs': self.docs }






"""
.. module:: riak_object.py

"""

import re
import csv
import urllib
from twisted.internet import defer

from riakasaurus import util, mapreduce, riak_link, riak_index_entry
from riakasaurus.metadata import *

class RiakObject(object):
    """
    The RiakObject holds meta information about a Riak object, plus the
    object's data.
    """

    def __init__(self, client, bucket, key=None):
        """
        Construct a new RiakObject.
        :param client: A RiakClient object.
        :param bucket: A RiakBucket object.
        :param key: Optional key.

        If key is not specified, then it is generated by server when
        store(...) is called.
        """
        self._client = client
        self._bucket = bucket
        self._key = key
        self._jsonize = True
        self._headers = {}
        self._links = []
        self._siblings = []
        self._metas = {}
        self._usermeta = {}
        self._indexes = []
        self._exists = False
        self._data = None

    def get_bucket(self):
        """
        Get the bucket of this object.
        @return RiakBucket
        .. todo:: remove accessor
        """
        return self._bucket

    def get_key(self):
        """
        Get the key of this object.
        @return string
        .. todo:: remove accessor
        """
        return self._key

    def get_data(self):
        """
        Get the data stored in this object. Will return a associative
        array, unless the object was constructed with new_binary(...) or
        get_binary(...), in which case this will return a string.
        @return array or string
        .. todo:: remove accessor
        """
        return self._data

    def get_encoded_data(self):
        """
        Get the data encoded for storing
        """
        if self._jsonize == True:
            content_type = self.get_content_type()
            encoder = self._client.get_encoder(content_type)
            if encoder is None:
                if isinstance(self._data, basestring):
                    return self._data.encode()
                else:
                    raise RiakError("No encoder for non-string data "
                                    "with content type ${0}".
                                    format(content_type))
            else:
                return encoder(self._data)
        else:
            return self._data

    def set_data(self, data):
        """
        Set the data stored in this object. This data will be
        JSON encoded unless the object was constructed with
        new_binary(...) or get_binary(...).
        @param mixed data - The data to store.
        @return data
        .. todo:: remove accessor
        """
        self._data = data
        return self

    def set_encoded_data(self, data):
        """
        Set the object data from an encoded string. Make sure
        the metadata has been set correctly first.
        """
        if self._jsonize == True:
            content_type = self.get_content_type()
            decoder = self._client.get_decoder(content_type)
            if decoder is None:
                # if no decoder, just set as string data for application to handle
                self._data = data
            else:
                self._data = decoder(data)
        else:
            self._data = data
        return self

    def status(self):
        """
        Get the HTTP status from the last operation on this object.
        @return integer
        .. todo:: convert to property
        """
        return self._headers['http_code']

    def exists(self):
        """
        Return True if the object exists, False otherwise. Allows you to
        detect a get(...) or get_binary(...) operation where the
        object is missing.
        @return boolean
        .. todo:: remove accessor
        """
        return self._exists

    def get_content_type(self):
        """
        Get the content type of this object. This is either text/json, or
        the provided content type if the object was created via
        new_binary(...).
        @return string
        .. todo:: convert to property
        """
        return self._headers.get('content-type', 'text/json')

    def set_content_type(self, content_type):
        """
        Set the content type of this object.
        @param string content_type - The new content type.
        @return self
        .. todo:: convert to property
        """
        self._headers['content-type'] = content_type
        return self

    def add_index(self, field, value):
        """
        Tag this object with the specified field/value pair for indexing.

        :param field: The index field.
        :type field: string
        :param value: The index value.
        :type value: string or integer
        :rtype: self
        """
        rie = riak_index_entry.RiakIndexEntry(field, value)
        if not rie in self._indexes:
            self._indexes.append(rie)

        return self

    def remove_index(self, field=None, value=None):
        """
        Remove the specified field/value pair as an index on this object.

        :param field: The index field.
        :type field: string
        :param value: The index value.
        :type value: string or integer
        :rtype: self
        """
        if not field and not value:
            ries = self._indexes[:]
        elif field and not value:
            ries = [x for x in self._indexes if x.get_field() == field]
        elif field and value:
            ries = [riak_index_entry.RiakIndexEntry(field, value)]
        else:
            raise Exception("Cannot pass value without a field name while removing index")

        for rie in ries:
            if rie in self._indexes:
                self._indexes.remove(rie)
        return self

    remove_indexes = remove_index

    def set_indexes(self, indexes):
        """
        Replaces all indexes on a Riak object. Currenly supports an iterable of 2 item tuples,
        (field, value)

        :param indexes: iterable of 2 item tuples consisting the field and value.
        :rtype: self
        """
        new_indexes = []
        for field, value in indexes:
            rie = riak_index_entry.RiakIndexEntry(field, value)
            new_indexes.append(rie)
        self._indexes = new_indexes

        return self

    def get_indexes(self, field = None):
        """
        Get a list of the index entries for this object. If a field is provided, returns a list

        :param field: The index field.
        :type field: string or None
        :rtype: (array of RiakIndexEntry) or (array of string or integer)
        """
        if field == None:
            return self._indexes
        else:
            return [x.get_value() for x in self._indexes if x.get_field() == field]

    def add_link(self, obj, tag=None):
        """
        Add a link to a RiakObject.
        :param obj:  RiakObject or a RiakLink object.
        :param tag: Optional link tag.
        :returns: RiakObject

        Default for the tag is the bucket name.
        Tag is ignored if obj is a RiakLink object.
        """
        if isinstance(obj, riak_link.RiakLink):
            newlink = obj
        else:
            newlink = riak_link.RiakLink(obj._bucket._name, obj._key, tag)

        self.remove_link(newlink)
        self._links.append(newlink)
        return self

    def remove_link(self, obj, tag=None):
        """
        Remove a link to a RiakObject.

        :param obj:RiakObject or a RiakLink object.
        :param tag: Optional link tag.
        :returns: self

        Default for the tag is the bucket name.
        Tag is ignored if obj is a RiakLink object.
        """
        if isinstance(obj, riak_link.RiakLink):
            oldlink = obj
        else:
            oldlink = riak_link.RiakLink(obj._bucket._name, obj._key, tag)

        a = []
        for link in self._links:
            if not link.isEqual(oldlink):
                a.append(link)

        self._links = a
        return self

    def get_links(self):
        """
        Return a list of RiakLink objects.

        :returns: list of RiakLink objects
        """
        # Set the clients before returning...
        for link in self._links:
            link._client = self._client
        return self._links

    def add_meta_data(self, key, data):
        """
        Add one meta data to object.
        Store in a dictionary so there's only one copy.

        :param key: string to add to metas dictionary
        :param data: data value for key
        :returns: self
        """
        self._metas[key] = data
        return self

    def remove_meta_data(self, data):
        """
        Remove meta data from object.

        :param data: string to remove from meta data.
        :returns: self
        """

        try:
            del(self._metas[data])
        except KeyError:
            pass

        return self

    def get_all_meta_data(self):
        """
        Return dictionary of meta data.
        """

        return self._metas

    @defer.inlineCallbacks
    def store(self, w=None, dw=None, pw=None, return_body=True, if_none_match=False):
        """
        Store the object in Riak. When this operation completes, the
        object could contain new metadata and possibly new data if Riak
        contains a newer version of the object according to the object's
        vector clock.

        :param w: Wait for this many partitions to respond
                  before returning to client.
        :param dw: Wait for this many partitions to confirm the
                   write before returning to client.
        :param return_body: If the newly stored object should be retrieved again.

        :returns: deferred

        """

        # Use defaults if not specified...
        w = self._bucket.get_w(w)
        dw = self._bucket.get_dw(dw)
        pw = self._bucket.get_pw(pw)

        # Issue the put over our transport
        t = self._client.transport

        if self._key is None:
            key, vclock, metadata = yield t.put_new(self, w=w, dw=dw, pw=pw, return_body=return_body, if_none_match=if_none_match)
            self._exists = True
            self._key = key
            self._vclock = vclock
            self.set_metadata(metadata)
        else:
            Result = yield t.put(self, w=w, dw=dw, pw=pw, return_body=return_body, if_none_match=if_none_match)
            if Result is not None:
                self.populate(Result)

        defer.returnValue(self)


    @defer.inlineCallbacks
    def reload(self, r=None, pr=None, vtag=None):
        """
        Reload the object from Riak. When this operation completes, the
        object could contain new metadata and a new value, if the object
        was updated in Riak since it was last retrieved.

        :param r: Wait for this many partitions to respond before
                  returning to client.
        :returns: self -- via deferred
        """
        # Do the request...
        r = self._bucket.get_r(r)

        r = self._bucket.get_r(r)
        pr = self._bucket.get_pr(pr)

        result = yield self._client.transport.get(self, r=r, pr=pr, vtag=vtag)

        self.clear()
        if result is not None:
            self.populate(result)

        defer.returnValue(self)

    @defer.inlineCallbacks
    def delete(self, w=None, dw=None):
        """
        Delete this object from Riak.

        :param w: wait for this many partitions to respond
            before returning to client.
        :type w: integer
        :param dw: wait for this many partitions to
            confirm the write before returning to client.
        :type dw: integer
        :returns: self -- via deferred
        """
        # Use defaults if not specified...
        rw = self._bucket.get_rw(rw)
        r = self._bucket.get_r(r)
        w = self._bucket.get_w(w)
        dw = self._bucket.get_dw(dw)
        pr = self._bucket.get_pr(pr)
        pw = self._bucket.get_pw(pw)

        # Run the operation...
        response = yield self._client.transport.delete(
                self, rw=rw, r=r, w=w, dw=dw, pr=pr, pw=pw)

        self.clear()
        defer.returnValue(self)

    def clear(self):
        """
        Reset this object.
        This is a local operation that clears the object,
        not the Riak stored version.
        :returns: self
        """
        self._headers = {}
        self._links = []
        self._data = None
        self._metas = {}
        self._usermeta = {}
        self._exists = False
        self._siblings = []
        self._indexes = []
        return self

    def vclock(self):
        """
        Get the vclock of this object.
        @return string
        .. todo:: convert to property
        """
        if ('x-riak-vclock' in self._headers.keys()):
            return self._headers['x-riak-vclock']
        else:
            return None

    def populate(self, Result) :
        """
        Populate the object based on the return from get.

        If None returned, then object is not found
        If a tuple of vclock, contents then one or more
        whole revisions of the key were found
        If a list of vtags is returned there are multiple
        sibling that need to be retrieved with get.
        """
        self.clear()
        if Result is None:
            return self
        elif type(Result) == types.ListType:
            self.set_siblings(Result)
        elif type(Result) == types.TupleType:
            (vclock, contents) = Result
            self._vclock = vclock
            if len(contents) > 0:
                (metadata, data) = contents.pop(0)
                self._exists = True
                if not metadata.has_key(MD_INDEX):
                    metadata[MD_INDEX] = []
                self.set_metadata(metadata)
                self.set_encoded_data(data)
                # Create objects for all siblings
                siblings = [self]
                for (metadata, data) in contents:
                    sibling = copy.copy(self)
                    sibling.set_metadata(metadata)
                    sibling.set_encoded_data(data)
                    siblings.append(sibling)
                for sibling in siblings:
                    sibling.set_siblings(siblings)
        else:
            raise RiakError("do not know how to handle type " + str(type(Result)))

    def set_metadata(self, metadata):
        """
        Set the metadata stored in this object.

        :param metadata: The data to store.
        :type metadata: dict
        :rtype: data
        """
        self._metas = metadata
        return self

    def get_metadata(self):
        """
        Get the metadata stored in this object. Will return an associative
        array

        :rtype: dict
        """
        return self._metas

    def get_usermeta(self):
        return self._usermeta

    def set_usermeta(self, usermeta):
        """
        Sets the custom user metadata on this object. This doesn't include things
        like content type and links, but only user-defined meta attributes stored
        with the Riak object.

        :param userdata: The user metadata to store.
        :type userdata: dict
        :rtype: data
        """
        self._usermeta = usermeta
        return self

    def _populate_metas(self):
        """
        Scan headers looking for x-riak-meta-
        """
        self._metas = {}
        for head in self._headers.keys():
            # Twisted lowercases headers.
            if 'x-riak-meta-' in head:
                data = self._headers[head]
                key = head.replace('x-riak-meta-', '')
                self._metas[key] = data

    def _populate_links(self, link_headers):
        """
        Parse out the links.

        :param link_headers: comma-delimited list of links
        :returns: self
        """
        # Twisted returns this as a list.
        for link_header in link_headers.strip().split(','):
            link_header = link_header.strip()
            matches = re.match(
                 "\<\/([^\/]+)\/([^\/]+)\/([^\/]+)\>; ?riaktag=\"([^\']+)\"",
                 link_header)
            if (matches != None):
                link = riak_link.RiakLink(urllib.unquote_plus(matches.group(2)),
                                          urllib.unquote_plus(matches.group(3)),
                                          urllib.unquote_plus(matches.group(4)))
                self._links.append(link)

        return self

    def has_siblings(self):
        """
        Return True if this object has siblings.

        """
        return(self.get_sibling_count() > 0)

    def get_sibling_count(self):
        """
        Get the number of siblings that this object contains.

        """
        return len(self._siblings)

    @defer.inlineCallbacks
    def get_sibling(self, i, r=None):
        """
        Retrieve a sibling by sibling number.

        :param i: sibling number
        :param r: R-Value. Wait until this many particitions have responded.
        :returns: RiakObject

        """
        # Use defaults if not specified.
        r = self._bucket.get_r(r)

        # Run the request...
        vtag = self._siblings[i]
        params = {'r': r, 'vtag': vtag}

        # Respond with a new object...
        obj = RiakObject(self._client, self._bucket, self._key)
        obj._jsonize = self._jsonize
        obj.populate(response)
        defer.returnValue(obj)

    @defer.inlineCallbacks
    def get_siblings(self, r=None):
        """
        Retrieve an array of siblings.

        :param r: R-Value. Wait until this many particitions have responded.
        :returns: array of RiakObject -- via deferred
        """
        a = []
        for i in range(self.get_sibling_count()):
            result = yield self.get_sibling(i, r)
            a.append(result)
        defer.returnValue(a)

    def add(self, *args):
        """
        Start assembling a Map/Reduce operation.
        see RiakMapReduce.add()

        :returns: RiakMapReduce
        """
        mr = mapreduce.RiakMapReduce(self._client)
        mr.add(self._bucket._name, self._key)
        return apply(mr.add, args)

    def link(self, *args):
        """
        Start assembling a Map/Reduce operation.
        see RiakMapReduce.link()
        :returns: RiakMapReduce
        """
        mr = mapreduce.RiakMapReduce(self._client)
        mr.add(self._bucket._name, self._key)
        return apply(mr.link, args)

    def map(self, *args):
        """
        Start assembling a Map/Reduce operation.
        see RiakMapReduce.map()
        :returns: RiakMapReduce
        """
        mr = mapreduce.RiakMapReduce(self._client)
        mr.add(self._bucket._name, self._key)
        return apply(mr.map, args)

    def reduce(self, *args):
        """
        Start assembling a Map/Reduce operation.
        see RiakMapReduce.reduce()
        :returns: RiakMapReduce
        """
        mr = mapreduce.RiakMapReduce(self._client)
        mr.add(self._bucket._name, self._key)

        return apply(mr.reduce, args)



"""
.. module:: client.py

RiakClient class

"""

import random
import base64
import urllib
import json
from twisted.internet import defer

from riakasaurus import mapreduce, util, bucket
from riakasaurus.search import RiakSearch

from riakasaurus import transport


class RiakClient(object):
    """
    The RiakClient object holds information necessary to connect to
    Riak.
    """
    def __init__(self, host='127.0.0.1', port=8098,
                prefix='riak', mapred_prefix='mapred',
                client_id=None, r_value="default", w_value="default", dw_value="default",
                transport=transport.HTTPTransport):
        """
        Construct a new RiakClient object.

        If a client_id is not provided, generate a random one.
        """
        self._host = host
        self._port = port
        self._prefix = prefix
        self._mapred_prefix = mapred_prefix
        if client_id:
            self._client_id = client_id
        else:
            self._client_id = 'py_' + base64.b64encode(
                                      str(random.randint(1, 1073741824)))
        self._r = r_value
        self._w = w_value
        self._dw = dw_value

        self._rw = "default"
        self._pr = "default"
        self._pw = "default"

        self._encoders = {'application/json': json.dumps,
                          'text/json': json.dumps}
        self._decoders = {'application/json': json.loads,
                          'text/json': json.loads}
        self._solr = None

        self.transport = transport(self) 

    def get_transport(self):
        return self.transport

    def get_r(self):
        """
        Get the R-value setting for this RiakClient. (default 2)
        :returns: integer representing current r-value
        .. todo:: remove accessor
        """
        return self._r

    def get_rw(self):
        """
        Get the RW-value for this ``RiakClient`` instance. (default "quorum")

        :rtype: integer
        """
        return self._rw

    def set_rw(self, rw):
        """
        Set the RW-value for this ``RiakClient`` instance. See :func:`set_r` for a
        description of how these values are used.

        :param rw: The RW value.
        :type rw: integer
        :rtype: self
        """
        self._rw = rw
        return self

    def get_pw(self):
        """
        Get the PW-value setting for this ``RiakClient``. (default 0)

        :rtype: integer
        """
        return self._pw

    def set_pw(self, pw):
        """
        Set the PW-value for this ``RiakClient`` instance. See :func:`set_r` for a
        description of how these values are used.

        :param pw: The W value.
        :type pw: integer
        :rtype: self
        """
        self._pw = pw
        return self

    def get_pr(self):
        """
        Get the PR-value setting for this ``RiakClient``. (default 0)

        :rtype: integer
        """
        return self._pr

    def set_pr(self, pr):
        """
        Set the PR-value for this ``RiakClient`` instance. See :func:`set_r` for a
        description of how these values are used.

        :param pr: The PR value.
        :type pr: integer
        :rtype: self
        """
        self._pr = pr
        return self

    def set_r(self, r):
        """
        Set the R-value for this RiakClient. This value will be used
        for any calls to get(...) or get_binary(...) where where 1) no
        R-value is specified in the method call and 2) no R-value has
        been set in the RiakBucket.
        @param integer r - The R value.
        @return self
        .. todo:: remove accessor
        """
        self._r = r
        return self

    def get_w(self):
        """
        Get the W-value setting for this RiakClient. (default 2)
        @return integer
        .. todo:: remove accessor
        """
        return self._w

    def set_w(self, w):
        """
        Set the W-value for this RiakClient. See set_r(...) for a
        description of how these values are used.
        @param integer w - The W value.
        @return self
        .. todo:: remove accessor
        """
        self._w = w
        return self

    def get_dw(self):
        """
        Get the DW-value for this ClientOBject. (default 2)
        @return integer
        .. todo:: remove accessor
        """
        return self._dw

    def set_dw(self, dw):
        """
        Set the DW-value for this RiakClient. See set_r(...) for a
        description of how these values are used.
        @param integer dw - The DW value.
        @return self
        .. todo:: remove accessor
        """
        self._dw = dw
        return self

    def get_client_id(self):
        """
        Get the client_id for this RiakClient.
        @return string
        .. todo:: remove accessor
        """
        return self._client_id

    def set_client_id(self, client_id):
        """
        Set the client_id for this RiakClient. Should not be called
        unless you know what you are doing.
        @param string client_id - The new client_id.
        @return self
        .. todo:: remove accessor
        """
        self._client_id = client_id
        return self

    def get_encoder(self, content_type):
        """
        Get the encoding function for the provided content type.
        """
        if content_type in self._encoders:
            return self._encoders[content_type]

    def set_encoder(self, content_type, encoder):
        """
        Set the encoding function for the provided content type.

        :param encoder:
        :type encoder: function
        """
        self._encoders[content_type] = encoder
        return self

    def get_decoder(self, content_type):
        """
        Get the decoding function for the provided content type.
        """
        if content_type in self._decoders:
            return self._decoders[content_type]

    def set_decoder(self, content_type, decoder):
        """
        Set the decoding function for the provided content type.

        :param decoder:
        :type decoder: function
        """
        self._decoders[content_type] = decoder
        return self

    def bucket(self, name):
        """
        Get the bucket by the specified name. Since buckets always exist,
        this will always return a RiakBucket.
        :returns: RiakBucket instance.
        """
        return bucket.RiakBucket(self, name)

    def is_alive(self):
        """
        Check if the Riak server for this RiakClient is alive.
        :returns: True if alive -- via deferred.
        """

        return self.transport.ping()

    def add(self, *args):
        """
        Start assembling a Map/Reduce operation.
        see RiakMapReduce.add()
        :returns: RiakMapReduce
        """
        mr = mapreduce.RiakMapReduce(self)
        return apply(mr.add, args)

    def search(self, *args):
        """
        Start assembling a Map/Reduce operation for Riak Search
        see RiakMapReduce.search()
        """
        mr = mapreduce.RiakMapReduce(self)
        return apply(mr.search, args)

    def link(self, args):
        """
        Start assembling a Map/Reduce operation.
        see RiakMapReduce.link()
        :returns: RiakMapReduce
        """
        mr = mapreduce.RiakMapReduce(self)
        return apply(mr.link, args)

    def list_buckets(self):
        """
        Retrieve a list of all buckets.

        :returns: list -- via deferred
        """

        return self.transport.get_buckets()

    def index(self, *args):
        """
        Start assembling a Map/Reduce operation based on secondary
        index query results.

        :rtype: :class:`RiakMapReduce`
        """
        mr = mapreduce.RiakMapReduce(self)
        return apply(mr.index, args)

    def map(self, *args):
        """
        Start assembling a Map/Reduce operation.
        see RiakMapReduce.map()
        :returns: RiakMapReduce
        """
        mr = mapreduce.RiakMapReduce(self)
        return apply(mr.map, args)

    def reduce(self, *args):
        """
        Start assembling a Map/Reduce operation.
        see RiakMapReduce.reduce()
        :returns: RiakMapReduce
        """
        mr = mapreduce.RiakMapReduce(self)
        return apply(mr.reduce, args)

    def solr(self):
        if self._solr is None:
            self._solr = RiakSearch(self, host=self._host, port=self._port)

        return self._solr


if __name__ == "__main__":
    pass



from xml.etree import ElementTree
from xml.dom.minidom import Document

from riakasaurus import util


class RiakSearch(object):
    def __init__(self, client, transport_class=None,
                 host="127.0.0.1", port=8098):
        self._client = client
        self._host = host
        self._port = port
        self._decoders = {"text/xml": ElementTree.fromstring}

    def get_decoder(self, content_type):
        decoder = self._client.get_decoder(content_type) or self._decoders.get(content_type)
        if not decoder:
            decoder = self.decode

        return decoder

    def decode(self, data):
        return data

    def _do_xml_request(self, url, xml):
        headers = {'Accept': 'text/xml, */*; q=0.5',
                   'Content-Type': 'text/xml',
                   }
        d = util.http_request_deferred('POST', self._host, self._port, url,
                                       headers, xml.toxml())
        d.addCallback(lambda response: response[0]['http_code'] == 200)
        return d

    def add(self, index, *docs):
        xml = Document()
        root = xml.createElement('add')
        for doc in docs:
            doc_element = xml.createElement('doc')
            for key, value in doc.iteritems():
                field = xml.createElement('field')
                field.setAttribute("name", key)
                text = xml.createTextNode(value)
                field.appendChild(text)
                doc_element.appendChild(field)
            root.appendChild(doc_element)
        xml.appendChild(root)
        return self._do_xml_request("/solr/%s/update" % index, xml)

    index = add

    def delete(self, index, docs=None, queries=None):
        xml = Document()
        root = xml.createElement('delete')
        if docs:
            for doc in docs:
                doc_element = xml.createElement('id')
                text = xml.createTextNode(doc)
                doc_element.appendChild(text)
                root.appendChild(doc_element)
        if queries:
            for query in queries:
                query_element = xml.createElement('query')
                text = xml.createTextNode(query)
                query_element.appendChild(text)
                root.appendChild(query_element)

        xml.appendChild(root)
        return self._do_xml_request("/solr/%s/update" % index, xml)

    remove = delete

    def search(self, index, query, **params):
        options = {'q': query, 'wt': 'json'}
        options.update(params)

        def decode_results(response):
            headers, data = response
            decoder = self.get_decoder(headers['content-type'])
            return decoder(data)

        url = "/solr/%s/select" % index
        host, port, url = util.build_rest_path(self._client, prefix=url,
                                               params=options)
        d = util.http_request_deferred('GET', host, port, url)
        d.addCallback(decode_results)
        return d

    select = search
